package loopbounds;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;
import org.jmlspecs.openjml.JmlTree;
import org.jmlspecs.openjml.JmlTree.JmlSpecificationCase;

import loopbounds.ng.*;
import loopbounds.parsetree.*;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCExpression;
import com.sun.tools.javac.tree.JCTree.JCMethodDecl;
import com.sun.tools.javac.tree.JCTree.JCStatement;


/**
 * Main class for the analysis of loop bounds. Can read a Java source file, has
 * some helper functions for selecting a loop to analyze, then when the
 * analyzeLoop method is called makes a hypothesis for a loop bound and outputs
 * JML annotated code, which can be verified with a tool like KeY.
 * 
 * @author Rody Kersten
 * @author Nick Gubbels
 */
public class LoopBoundsAnalyzer {

	// The GUI may change this var, according to the properties file
	public static boolean DEBUG = false;

	private TestingUnit tu = null;
	private PolynomialInterpolator pg = null;
	private StringWriter diagnostic;
	private String[] classPath = new String[]{};
	
	/** Default constructor.
	 * Diagnostic output is printed to the 'diagnostic' StringWriter, which can
	 * be forwarded (for instance) to the console.
	 **/	
	public LoopBoundsAnalyzer(StringWriter diagnostic) throws java.io.IOException {
		this.diagnostic = diagnostic;
	}

	/**
	 * Get the value of the 'diagnostic' string buffer, where information about
	 * the inference process is written to. Can be read by a user interface when
	 * it wants to output this information.
	 * 
	 * @return The value of the 'diagnostic' string buffer
	 */
	public String getDiagnostic() {
		String res = diagnostic.toString();
		StringBuffer buf = diagnostic.getBuffer();
		buf.setLength(0);
		return res;
	}
	
	/**
	 * Analyze a loop, i.e. suggest a piecewise polynomial bound for this loop,
	 * given an expected degree of the polynomial.
	 * 
	 * @param degree
	 *            The expected degree of the polynomial.
	 * @param step
	 * 			A fixed step. If <= 0, the analyzer will try to
	 * 			find the steps automatically
	 * @param splitIfs
	 * 			Boolean which indicates if branch-splitting should be applied
  	 * @param correctJML
	 * 			  If true, write JML annotations that are correct with respect to the
	 * 			  JML specification, rather than annotations that can be verified by KeY.
	 * @param overwrite
	 * 			Boolean which indicates if an existing decreases clause should be overwritten
	 * @param nodeStartPoint
	 * 			Startpoint for nodesearch
	 * @param nodeMinVal
	 * 			Minimal value for nodes
	 * @param nodeMaxVal
	 * 			Maximal value for nodes
	 * @return The adapted and annotated source code as can be verified by the
	 *         KeY tool.
	 * @throws Exception
	 */
	public String analyseAllLoops(ParseTreeWrapper input, int degree,
			int step, boolean splitIfs, boolean correctJML, boolean overwrite,
			int nodeStartPoint,	int nodeMinVal, int nodeMaxVal, boolean refactoring)
			throws UndeclaredVariableException,	UnsupportedTypeException, Exception {
	
		diagnostic.write("---- ANALYSING ALL LOOPS ----\n\n");
		
		for (int c = 0; c < input.getNumClasses(); c++) {
			for (int m = 0; m < input.getNumMethods(c); m++) {
				for (int l = 0; l < input.getNumLoops(c,m); l++) {
					String newcode =
						analyseLoop(input, c, m, l, degree, step, splitIfs, correctJML, 
								overwrite, nodeStartPoint, nodeMinVal, nodeMaxVal, refactoring);
					input.parseSource(newcode);
				}
			}
		}
		
		return input.getSource();
	}


	/**
	 * Analyze a loop, i.e. suggest a piecewise polynomial bound for this loop,
	 * given an expected degree of the polynomial.
	 * 
	 * @param input
	 * 			  The input parse tree
	 * @param classIndex
	 *            The index of the class in which the loop to analyze is.
	 * @param methIndex
	 *            The index of the method in which the loop to analyze is.
	 * @param loopIndex
	 *            The index of the loop to analyze, within the considered
	 *            method.
	 * @param degree
	 *            The expected degree of the polynomial.
	 * @param step
	 * 			A fixed step. If <= 0, the analyzer will try to
	 * 			find the steps automatically
	 * @param splitIfs
	 * 			Boolean which indicates if branch-splitting should be applied
	 * @param correctJML
	 * 			  If true, write JML annotations that are correct with respect to the
	 * 			  JML specification, rather than annotations that can be verified by KeY.
	 * @param overwrite
	 * 			Boolean which indicates if an existing decreases clause should be overwritten
	 * @param nodeStartPoint
	 * 			Startpoint for nodesearch
	 * @param nodeMinVal
	 * 			Minimal value for nodes
	 * @param nodeMaxVal
	 * 			Maximal value for nodes
	 * @return The adapted and annotated source code as can be verified by the
	 *         KeY tool.
	 * @throws Exception
	 */
	public String analyseLoop(ParseTreeWrapper input, int classIndex, int methIndex, int loopIndex, int degree,
			int step, boolean splitIfs, boolean correctJML, boolean overwrite, int nodeStartPoint, int nodeMinVal, int nodeMaxVal, boolean refactoring)
			throws UndeclaredVariableException,	UnsupportedTypeException, Exception {
	
		diagnostic.write("******** STARTING ANALYSIS ********\n\n");

		// record time elapsed in this method
		long startTime = System.currentTimeMillis();
		
		JCStatement loop = input.getLoop(classIndex, methIndex, loopIndex);
		
		//DNF-split
		DNFSplitter ds = new DNFSplitter(input);
		List<JCStatement> loops = ds.split(loop);

		// Output number of pieces in piecewise polynomial
		diagnostic.write("The piecewise polynomial has " + loops.size()
				+ " piece(s) for which the bound is non-zero.\n");
		diagnostic.write("Generating polynomial for each piece...\n\n");
		diagnostic.write("----------------------------------------------------------------------------------------------------------------------\n");

		// For each piece, calculate the polynomial
		int numNotFound = 0;
		PiecewisePolynomial lbf = new PiecewisePolynomial();
		for (JCStatement curloop : loops) {
			JCExpression condition = ParseTreeWrapper.getLoopCondition(curloop);
			Polynomial p = analysePiece(
					input,
					(JCExpression) condition.clone(), //copy because methods are augmented in place
					classIndex,
					methIndex,
					loopIndex,
					degree,
					step,
					splitIfs,
					nodeStartPoint,
					nodeMinVal, 
					nodeMaxVal,
					refactoring
					);
			if (!p.isUndefined())
				lbf.add(condition, p);
			else numNotFound++;
		}
		assert (loops.size() == lbf.numPieces());
		if (DEBUG)
			System.out.println("Found decreases clause: " + lbf.asDecreasesClause(input.context()).toString());
		
		// output elapsed time
		long stopTime = System.currentTimeMillis();
		long elapsedTime = stopTime - startTime;
		diagnostic.write("Calculated LBFs in " + elapsedTime + " ms.\n\n");

		if (numNotFound > 0) {
			diagnostic.write("Unfortunately, polynomials could not be found for "
					+ numNotFound + " pieces\n" + "You have to either specify" +
						" these by hand or try a higher degree.\n\n");
		}	
		
		// load a copy of the AST
		ParseTreeWrapper output = input.freshcopy();

		// annotate
		boolean annotated = AnnotationGenerator.annotate(
				lbf,
				output.getMethod(classIndex, methIndex),
				output.getLoop(classIndex, methIndex, loopIndex),
				output.context(),
				correctJML,
				overwrite
			);
		
		String outsource = input.getSource();
		
		if (annotated) {
			diagnostic.write("The loop was annotated successfully, you can now try to" +
					" prove the inferred decreases clause in KeY, by verifying" +
					" the 'ensuresPost' proof obligation.\n\n");
			
			com.sun.tools.javac.util.List<JmlTree.JmlStatementLoop> loopAnn =
				ParseTreeWrapper.getLoopSpecs(output.getLoop(classIndex, methIndex, loopIndex));
			com.sun.tools.javac.util.List<JmlSpecificationCase> methAnn =
				output.getMethod(classIndex, methIndex).cases.cases;
			outsource = input.replaceAnnotations(outsource, classIndex, methIndex, loopIndex, methAnn, loopAnn);
		} else {
			diagnostic.write("[WARNING] The loop was already annotated with a decreases-clause," +
					" which was not overwritten.\n\n");
		}

		return outsource;
	}

	/** Does the analysis for a single piece, defined by condition 'condition' */
	private Polynomial analysePiece(ParseTreeWrapper input, JCExpression condition,
			int classIndex,	int methIndex, int loopIndex,
			int degree, int step, boolean splitIfs,
			int nodeStartPoint, int nodeMinVal, int nodeMaxVal, boolean refactoring
			) throws Exception {

		diagnostic.write("Generating polynomial for piece with condition: "
				+ condition + "\n\n");

		// Prepare output tree and get the loop from this one
		ParseTreeWrapper copy = input.freshcopy();
		JCStatement loopCopy = copy.getLoop(classIndex, methIndex, loopIndex);
		
		//Modify the loop by setting the condition obtained by DNF-splitting
		ParseTreeWrapper.setLoopCondition(loopCopy, condition);
		ParseTreeWrapper.removeForLoopInit(loopCopy);
	
		// Split branches, if necessary
		Polynomial p;
		if (splitIfs) {
			
			//Split the branches
			List<JCStatement> loopsAfterSplit = IfSplitter.split(loopCopy, copy.context());
			diagnostic.write("Splitting branches, there are " + loopsAfterSplit.size() + "\n\n");
			List<Polynomial> ps = new ArrayList<Polynomial>();
			
			//Do the analysis for each branch
			for (JCStatement loop : loopsAfterSplit) {
				ps.add(calcPolynomial(classIndex, methIndex, loopIndex,
						degree, step, copy, loop, nodeStartPoint, nodeMinVal, nodeMaxVal, refactoring));
			}
			
			// Combine the polynomials
			p = Polynomial.bound(ps);
			diagnostic.write("Found worst case polynomial: " + p + "\n");

		} else {
			// No branch-splitting: do the analysis for the single branch
			p = calcPolynomial(classIndex, methIndex, loopIndex, degree,
					step, copy, loopCopy, nodeStartPoint, nodeMinVal, nodeMaxVal, refactoring);
		}
		
		if (p.isUndefined()) {
			diagnostic.write("Can't find polynomial of degree " + degree
					+ ", for condition " + condition + "\n\n");
			
		} else if (p.getStep() > 1) {
			// ceil polynomial if necessary
			p.ceil();
			diagnostic.write("Ceiled to: " + p + "\n");
		}
		diagnostic.write("----------------------------------------------------------------------------------------------------------------------\n");
		return p;
	}

	/** Do the analysis for a single branch */
	private Polynomial calcPolynomial(int classIndex, int methIndex, int loopIndex,
			int degree, int step, ParseTreeWrapper input, JCStatement loopCopy,
			int nodeStartPoint, int nodeMinVal, int nodeMaxVal, boolean refactoring)
			throws Exception {
		
		//Make a copy of the tree, which will be instrumented with a counter
		ParseTreeWrapper instrumented = input.freshcopy();
		
		// Do simple slicing
		//Slicer.slice(loopCopy, instrumented, classIndex, methIndex);
		
		// Check if any objects are used in the loop, which are not declared or static
		ObjectSet usedObjects = ObjectScanner.getUsedObjects(loopCopy);
		usedObjects.removeStaticObjects();
		usedObjects.removeAll(ObjectScanner.getDeclaredObjects(loopCopy));
		
		// get undeclared variables
		VarInfo variables = VariableScanner.getUndeclaredVars(loopCopy);
		variables.rename(instrumented, classIndex, methIndex, loopIndex);
		usedObjects.addAll(variables.getRefactoredObjects());
		usedObjects.removeAll(variables.getFullNames());
	
		//Replace variables by their placeholders
		VariableReplacer.replaceVariables(loopCopy, variables, instrumented.context());
		
		// Create new method
		JCMethodDecl newMethod = MethodCreator.createMethod(
				loopCopy,
				instrumented,
				classIndex,
				methIndex,
				variables,
				true
				);
		
		if (refactoring) {
			// Remove statements with locally undeclared objects and give
			// a warning
			List<JCTree> removedStatements = ObjectRemover
					.removeObjectStatements(newMethod, usedObjects);

			if (!removedStatements.isEmpty()) {
				diagnostic
						.write("WARNING - following statement(s) are removed by automated refactoring:\n");
				for (JCTree statement : removedStatements)
					diagnostic.write("  " + statement.toString() + "\n");
				diagnostic
						.write("because following object(s) were unkown in the generated augmented method\n");
				for (String s : usedObjects)
					diagnostic.write("  " + s + "\n");
				diagnostic.write("\n");
			}
		}
		
		//Add new method to the output tree
		MethodAdder.addSingle(newMethod, instrumented.getInputTree(), classIndex);
		diagnostic.write("Generated augmented method:\n" + newMethod + "\n\n");

		// Compile the output tree
		String instrumentedCode = instrumented.prettyPrint();
		
		String packageName = "";
		if (instrumented.getClass(classIndex).toplevel.pid != null) {
		    packageName = instrumented.getClass(classIndex).toplevel.pid.toString(); 
		}
		
		// Compile the code and create the testing unit
		try {
			tu = new TestingUnit(packageName,
					instrumented.getClassName(classIndex),
					newMethod.name.toString(), variables, instrumentedCode,
					classPath);
		} catch (CompilationException e) {
			diagnostic.write("ERROR - there were compilation errors during testing:\n");
			diagnostic.write(e.getErrorString());
			return Polynomial.undefined();
		}

		
		//Infer the step, if necessary
		if (step <= 0)
			step = StepFinder.getStep(loopCopy);

		// Generate test nodes
		NodeGenerator ng = new SimpleNodeGenerator(
				degree,
				step,
				variables.getPlaceholders().toArray(new String[0]),
				variables.getTypes().toArray(new Class<?>[0]),
				ParseTreeWrapper.getLoopCondition(loopCopy),
				nodeStartPoint, //startPoint
				nodeMinVal, //minval
				nodeMaxVal //maxval
				);
		try {
			ng.generate();
		} catch (NotEnoughTestNodesException e) {
			diagnostic.write(e.toString() + "\n\n");
			return Polynomial.undefined();
		}
		List<int[]> testnodes = ng.getNodes();
		
		// run the tests
		int[] results = tu.doTests(testnodes);

		// interpolate
		pg = new PolynomialInterpolator(
				variables,
				degree,
				testnodes,
				results);
		Polynomial p = pg.generate(new PrintWriter(diagnostic));
		
		//Apply step
		if (step > 0) {
			p.applyStep(step);
		}
		
		diagnostic.write("Step: " + step + "\n\n");
		diagnostic.write("Found polynomial: " + p + "\n\n");
		return p;
	}
	
	public void setClasspath(String[] classpath){
		classPath = classpath;
	}
}
