package loopbounds;

import java.util.ArrayList;
import java.util.List;

import org.jmlspecs.openjml.JmlTree;

import Jama.Matrix;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCExpression;
import com.sun.tools.javac.tree.JCTree.JCIdent;
import com.sun.tools.javac.util.Context;

/**
 * Class for storing a polynomial, with some functions to pretty print it.
 * 
 * The polynomial is stored as:
 * <ul>
 * <li>A List of variables</li>
 * <li>A List of Lists of exponents (integers). Each list has the same length as
 * the number of variables.</li>
 * <li>A Matrix of coefficients (doubles).</li>
 * <li>A step</li>
 * </ul>
 * 
 * @author Rody Kersten
 * 
 */
public class Polynomial {
	private VarInfo variables;
	private List<List<Integer>> exponents;
	private Matrix coefficients;
	private int step;

	/**
	 * Default constructor.
	 * 
	 * @param variables
	 *           The variables of the polynomial
	 * @param exponents
	 *            List of Lists of exponents (integers). Each list has the same
	 *            length as the number of variables.
	 * @param coefficients
	 *            A Matrix of coefficients (doubles)
	 * @param step
	 *            The step size, meaning the coefficients are divided by 'step'
	 *            to create rational coefficients
	 */
	public Polynomial(VarInfo variables, List<List<Integer>> exponents,
			Matrix coefficients, int step) {
		this.exponents = exponents;
		this.coefficients = coefficients;
		this.step = step;
		this.variables = variables;
	}

	/**
	 * Default constructor.
	 * 
	 * @param variables
	 *            The variables of the polynomial
	 * @param exponents
	 *            List of Lists of exponents (integers). Each list has the same
	 *            length as the number of variables.
	 * @param coefficients
	 *            A Matrix of coefficients (doubles)
	 */
	public Polynomial(VarInfo variables, List<List<Integer>> exponents,
			Matrix coefficients) {
		this(variables, exponents, coefficients, 1);
	}

	/**
	 * Get the variables in this polynomial.
	 * 
	 * @return The variables in this polynomial
	 */
	public VarInfo getVariables() {
		return variables;
	}
	
	/**
	 * Gets the step variable.
	 * 
	 * @return The step.
	 */
	public int getStep() {
		return step;
	}

	/**
	 * If we can't find a polynomial we set it to 'undefined', which is equal to
	 * '-1'.
	 * 
	 * @return A polynomial representing 'undefined'.
	 */
	public static Polynomial undefined() {
		ArrayList<List<Integer>> exp = new ArrayList<List<Integer>>();
		exp.add(new ArrayList<Integer>());
		return new Polynomial(null,	exp, new Matrix(1, 1, -1));
	}

	/**
	 * Check if this polynomial represents 'undefined'.
	 * 
	 * @return true, if the polynomial is undefined, false otherwise.
	 */
	public boolean isUndefined() {
		return (variables == null) && (exponents.size() == 1) && (exponents.get(0).size() == 0)
				&& (coefficients.get(0, 0) == -1);
	}

	/**
	 * Export the polynomial to it's naive String representation.
	 * 
	 * @return A String representation of the polynomial. Not pretty printed,
	 *         for instance 0.0 * a + -1.0 * b
	 */
	public String toString() {

		if (isUndefined()) {
			return new String("UNDEFINED");
		}

		String p = new String();

		for (int i = 0; i < exponents.size(); i++) {
			p = p.concat(String.valueOf(coefficients.get(i, 0)));
			for (int j = 0; j < exponents.get(i).size(); j++) {
				if (exponents.get(i).get(j) != 0) {
					p = p.concat("(" + variables.getFullName(j) + "^"
							+ exponents.get(i).get(j) + ")");
				}
			}
			if (i < exponents.size() - 1)
				p = p.concat(" + ");
		}

		if (step > 1)
			return "(" + p + ") / " + step;
		return p;
	}

	/**
	 * Build a JCExpression containing the polynomial. Needed for pretty
	 * printing JML.
	 * 
	 * @return A JCExpression containing the polynomial in a nice form, i.e. not
	 *         containing terms that are multiplied with 0, additions of 0 or
	 *         multiplication with 1.
	 */
	public JCExpression toJCExpression(Context context) {
		JmlTree.Maker M = JmlTree.Maker.instance(context);

		// Exceptional case, where polynomial is not found so '-1' is returned
		if (isUndefined()) {
			return M.Literal(-1);
		}

		List<JCExpression> plus = new ArrayList<JCExpression>();
		List<JCExpression> min = new ArrayList<JCExpression>();

		for (int i = 0; i < exponents.size(); i++) {

			// ceil double to obtain integer version of coefficient
			// ceil, because polynomial provides an UPPER bound
			int coef = (int) Math.ceil(coefficients.get(i, 0));

			// if coef==0, then don't add it to p
			if (coef != 0) {

				// Convert coefficient to KeY tool readable JML
				JCExpression term = null;

				// Don't add multiplications with 1
				if ((coef < -1) || ((coef == -1) && (i == 0))) {
					term = M.Literal(Math.abs(coef));
				} else if ((coef > 1) || ((coef == 1) && (i == 0))) {
					term = M.Literal(coef);
				}

				// Build term
				for (int j = 0; j < exponents.get(i).size(); j++) {
					JCIdent var = M.Ident(variables.getFullName(j));
					for (int k = 0; k < exponents.get(i).get(j); k++) {
						if (term == null)
							term = var;
						else
							term = M.Binary(JCTree.MUL, term, var);
					}
				}

				// add term to total polynomial
				if (coef > 0) {
					plus.add(term);
				} else {
					min.add(term);
				}
			}
		}

		//Build polynomial from plus and min parts
		JCExpression p;
		if (plus.size() != 0) {
			p = plus.get(0);
			for (int i = 1; i < plus.size(); i++) {
				p = M.Binary(JCTree.PLUS, p, plus.get(i));
			}
			for (JCExpression term : min) {
				p = M.Binary(JCTree.MINUS, p, term);
			}
		} else if (min.size() != 0){
			p = M.Unary(JCTree.NEG, min.get(0));
			for (int i = 1; i < min.size(); i++) {
				p = M.Binary(JCTree.MINUS, p, min.get(i));
			}
		} else {
			return M.Literal(0);
		}

		if (step > 1)
			return M.Binary(JCTree.DIV, p, M.Literal(step));
		return p;
	}

	/**
	 * The prototype can infer polynomial loop bounds with floating point
	 * coefficients. However, only loop bounds with integer coefficients can be
	 * expressed in the current JML syntax and semantics.
	 * 
	 * As an intermediate fix we can handle rational coefficients (which are
	 * important when a variable on which the loop condition depends is
	 * increased with more than 1 each iteration, see example), by trying to
	 * find a common denominator for all coefficients, then using this as a step
	 * size. With this step size we mean that the decreases clause decreases on
	 * every iteration not by 1, but by 'step'. This means that if our
	 * polynomial p can be written as q/step, we verify that q decreases on
	 * every iteration, namely by step.
	 * 
	 * An example of a loop that can be verified thanks to this fix:
	 * 
	 * <pre>
	 * while (i &lt; 100) {
	 * 	i += 3;
	 * }
	 * </pre>
	 * 
	 * @return An integer which is a denominator for all coefficients, given
	 *         these are rationals. When any of the coefficients is multiplied
	 *         with this number the result is an integer. When such a
	 *         denominator cannot be found, 1 is returned.
	 */
	public int findDenominator() {
		for (int i = 1; i < 1000; i++) {
			if (isDenom(i, 0.0))
				return i;
		}
		return 1;
	}

	private boolean isDenom(int i, double maxdev) {
		for (double coef : coefficients.getRowPackedCopy()) {
			double prod = coef * i;
			if (!isInteger(prod, maxdev))
				return false;
		}
		return true;
	}

	/**
	 * Checks if the supplied step can be used to write this polynomial as
	 * p/step.
	 * 
	 * @param step
	 *            The supposed denominator
	 * @return True, if 'step' is a denominator of all coefficients
	 */
	public boolean checkStep(int step) {
		return isDenom(step, 0.01);
	}

	private boolean isInteger(double d, double maxdev) {
		return (Math.abs(Math.round(d) - d) <= maxdev);
		// return (Math.round(d) == d);
	}

	/**
	 * Checks if all coefficients are integers (i.e. have no numbers behind the
	 * comma)
	 * 
	 * @return True, if all the coefficients of this polynomial are integers.
	 */
	public boolean allIntegerCoefficients() {
		if (step != 1)
			return false;
		for (double coef : coefficients.getRowPackedCopy()) {
			if (!isInteger(coef, 0.0))
				return false;
		}
		return true;
	}

	/**
	 * Applies a found 'step' to the polynomial. See documentation of
	 * findDenominator for what this step is and how it is found.
	 * 
	 * Multiplies all coefficients by 'step' and sets 'this.step' to 'step'.
	 * 
	 * @param step
	 *            The step to apply.
	 */
	public void applyStep(int step) {
		this.step = step;
		coefficients = coefficients.times(step);
	}
	
	/**
	 * Applies provisional ceiling. Since JML does not have a ceiling operation,
	 * this method can be used to add 'step-1' to the polynomial if the coefficients are
	 * all rationals with 'step' as denominator. This bounds the ceiling operation.
	 */
	public void ceil() {
		if (step > 1) {
			add(step-1);
		}
	}

	/**
	 * Adds n to the polynomial (by adding it to the first coefficient).
	 */
	public void add(int n) {
		coefficients.set(0, 0, coefficients.get(0, 0) + n);
	}

	/**
	 * Maximize the current polynomial with the one given as a parameter
	 * 
	 * @param Polynomial
	 *            p
	 */
	public void maximizePoly(Polynomial p) {
		for (int j = 0; j < coefficients.getColumnDimension(); j++) {
			for (int i = 0; i < coefficients.getRowDimension(); i++) {
				if (p.coefficients.get(i, j) > coefficients.get(i, j)) {
					coefficients.set(i, j, p.coefficients.get(i, j));
				}
			}
		}
	}
	
	/** Check if p1 and p2 share variables */
	private static boolean shareVars(Polynomial p1, Polynomial p2) {
		for (String var1 : p1.getVariables().getFullNames())
			for (String var2 : p2.getVariables().getFullNames())
					if (var1.equals(var2))
						return true;
		return false;
	}
	
	/** Evaluate this polynomial with all variables set to the value 'val' */
	private double evaluateWithEqualValues(double val) {
		
		if (isUndefined())
			return -1;
		
		double result = 0.0;

		for (int i = 0; i < exponents.size(); i++)
			for (int j = 0; j < exponents.get(i).size(); j++)
				result += coefficients.get(i, 0) * Math.pow(val,exponents.get(i).get(j));

		if (step > 1)
			result /= step;
		return result;
	}
	
	/** Compare a list polynomials and devise a polynomial which is an upper bound to all.
	 * 
	 * Used when branch-splitting is applied and thus multiple polynomials are inferred.
	 * 
	 * If the polynomials share variables, use the one that evaluates to higher results
	 * on equal values (in agreement to the worst-case-branch criterion).
	 * 
	 * If the polynomials are over different variables, add them. This is a slight addition
	 * to include also some cases that do not satisfy the worst-case-branch criterion.
	 * 
	 * @param ps A list of polynomials
	 * @return A polynomial which is an upper bound to the polynomials in the list
	 */
	public static Polynomial bound(List<Polynomial> ps) {
		
		//skip to the first defined polynomial
		int i = 0;
		while (i < ps.size() && ps.get(i).isUndefined()) i++;
		if (i >= ps.size()) return Polynomial.undefined();
		
		if (LoopBoundsAnalyzer.DEBUG) {
			System.out.print("Comparing polynomials: ");
			for (Polynomial p : ps)
				System.out.print(p + ", ");
			System.out.print("\n");
		}
		
		Polynomial bnd = ps.get(i);
		i++;
		
		while(i < ps.size()) {
			Polynomial pi = ps.get(i);
			if (!pi.isUndefined()) {

				if (shareVars(bnd,pi)) {
					//if bnd and pi share variables, then use max for all vars set
					// to an equal number, try 1-9, then conclude polynomials are equal
					double evalpi;
					double evalbnd;
					int val = 1;
					do {
						evalpi = pi.evaluateWithEqualValues(val);
						evalbnd = bnd.evaluateWithEqualValues(val);
						if (LoopBoundsAnalyzer.DEBUG) {
							System.out.println("Eval of "+pi+" = " + evalpi);
							System.out.println("Eval of "+bnd+" = " + evalbnd);
						}
						val++;
					} while (evalpi == evalbnd && pi.step == bnd.step && val < 10);
					if (evalpi > evalbnd || (evalpi==evalbnd && pi.step < bnd.step)) {
						bnd = pi;
					}
				} else {
					//if bnd and pi do not share variables, add them
					for (int a = 0; a < bnd.coefficients.getRowDimension(); a++) {
						for (int b = 0; b < bnd.coefficients.getColumnDimension(); b++) {
							if (pi.coefficients.get(a, b) > bnd.coefficients.get(a, b)) {
								bnd.coefficients.set(a, b, pi.coefficients.get(a, b));
							}
						}
					}
				}
			}
			i++;
		}
		
		return bnd;
	}
}
