package loopbounds;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

import Jama.Matrix;

/**
 * Class for generating a polynomial loop bound. Uses the Jama package
 * to find the unique polynomial interpolation for a set of test nodes and
 * their results.
 * 
 * @author Martijn Moraal
 * @author Rody Kersten
 *
 */
public class PolynomialInterpolator {
	VarInfo variables;
	List<List<Integer>> exponents;
	int degree;
	List<int[]> testnodes;
	int[] results;
		
	/**
	 * Constructor for the PolynomialInterpolator class. Sets all the required data
	 * for interpolation. Follow by a call to 'generate' to do the actual interpolation.
	 * 
	 * @param variables	The variables on which the loop bound depends, the order of these names needs to correspond to those in the CompiledUnit.
	 * @param degree	The expected degree of the polynomial.
	 * @param testnodes	The nodes on which the tests were conducted.
	 * @param results	The test results. The value of testresults[i] should be the test result for input node testnodes.get(i).
	 */
	public PolynomialInterpolator (VarInfo variables, int degree, List<int[]> testnodes, int[] results) {

		this.variables = variables;
		this.degree = degree;
		this.testnodes = testnodes;
		this.results = results;
		
		//Special case: if there are 0 variables, the max degree is 0
		if (variables.size() == 0)
			this.degree = 0;
		
		//List of exponents, these are static and only dependent on the amount of variables and degree
		this.exponents = generate_tuples(variables.size(), this.degree);
	}
	
	/** Generate the polynomial, while outputting the test nodes, results and
	 * found coefficients.
	 * 
	 * @param pw PrintWriter to write output to
	 * @return The polynomial interpolation.
	 **/
	public Polynomial generate (PrintWriter pw) {
				
		if (results.length == 0 && variables.size() > 0)
				return Polynomial.undefined();
		
		//Convert nodes and results to proper format
		double[][] res = new double[testnodes.size()][1];
		int[][] nodes = new int[testnodes.size()][variables.size()];
		double[][] dnodes = new double[testnodes.size()][variables.size()];
		for (int i = 0; i < testnodes.size(); i++) {
			res[i][0] = results[i];
			nodes[i] = testnodes.get(i);
			for (int j = 0; j < variables.size(); j++)
				dnodes[i][j] = (double)testnodes.get(i)[j];
		}
		
		//Build matrix
		double[][] matrix = buildMatrix(variables.size(), degree, nodes);
		
		Matrix A = new Matrix(res);
		Matrix B = new Matrix(matrix);
		Matrix nodesMatrix = new Matrix(dnodes);
		
		//Solve matrix
		Matrix coefficients = B.solve(A);
		
		//Output to console
		if (pw != null) {
			pw.print("Input values:\n");
			nodesMatrix.print(pw, 8, 0);
			pw.print("Count:\n");
			A.print(pw, 8, 0);
		}
		
		return new Polynomial(variables, exponents, coefficients);
	}
	
	/** Generate the polynomial, without outputting anything.
	 * @return The polynomial interpolation.
	 * */
	public Polynomial generate() {
		return generate(null);
	}
	
	private double[][] buildMatrix (int n, int d, int[][] nodes) {
		//Builds the matrix for the polynomial interpolation
		
		//The amount of test nodes must be equal to the amount of the coefficients
		assert(nodes.length == binom(d+n, n));
		
		double[][] matrix = new double[nodes.length][];
		
		for (int i = 0; i < nodes.length; i++) {
			matrix[i] = buildRow(n, d, nodes[i]);
		}
		
		return matrix;
	}
	
	private double[] buildRow (int n, int d, int[] node) {
		//Builds a single row for the matrix
		// d -- degree of the polynomial to reconstruct
		// n -- the amount of variables
		// node -- a test node (without the corr. result)
		assert(n == node.length);
		
		double[] row = new double[exponents.size()];
		
		for (int i = 0; i < exponents.size(); i++) {
			List<Integer> current_degrees = exponents.get(i);
			double prod = 1;
			//X1^(i1) ... Xn^(in)
			for (int j=0; j < n; j++) {
				prod *= Math.pow(node[j], current_degrees.get(j));
			}
			row[i] = prod;
		}
		return row;
	}
	
	private static List<List<Integer>> generate_tuples (int n, int d) {
		//generates the n-tuple of integers, the sum of which is at most d
		int length = binom(d+n, n);
		
		List<List<Integer>> result = new ArrayList<List<Integer>>();
		
		for (int k = 0; k <= d; k++) {
			List<List<Integer>> intermediate_result = generate_tuples_aux(n, k);
			result.addAll(intermediate_result);
		}
		
		assert(result.size() == length);
		
		return result;
	}
	
	private static List<List<Integer>> generate_tuples_aux (int n, int k) {
		//generates the n-tuple of integers, the sum of which is exactly k
		List<List<Integer>> result = new ArrayList<List<Integer>>();
		
		if (n <= 1) {
			List<Integer> tuple = new ArrayList<Integer>();
			if (n==1)
				tuple.add(k);
			result.add(tuple);
		} else {
			for (int i = k; i >= 0; i--) {
				List<List<Integer>> intermediate_tuples = generate_tuples_aux(n-1, i);
				
				for(int j = 0; j < intermediate_tuples.size(); j++) {
					intermediate_tuples.get(j).add(k-i);
				}
				result.addAll(intermediate_tuples);
			}
		}
		return result;
	}
	
	private static int binom (int n, int r) {
		// returns (n!/(r!(n-r)!))
		int t = 1;
		
		int m = n - r;
		if (r < m) {
			r = m;
		}
		
		for (int i = n, j = 1; i > r; i--, j++) {
			t = t * i / j;
		}
		
		return t;
	}
}
