package loopbounds;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.Array;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.List;

import loopbounds.parsetree.TypeScanner;
import loopbounds.parsetree.UnsupportedTypeException;

import com.sun.tools.javac.Main;
import com.sun.tools.javac.tree.JCTree.JCArrayTypeTree;
import com.sun.tools.javac.tree.JCTree.JCExpression;

/**
 * Class that conducts a series of tests. It runs a given method on each
 * node in a given set of test nodes and stores the (integer) return values.
 * 
 * It runs the method in a separate thread to ensure termination of
 * the prototype, even when the method would not normally terminate.
 * When the thread has not finished after approximately 16 seconds
 * it is assumed to run indefinitely, and it is killed.
 * An IndefiniteLoopException is thrown to indicate this.
 * 
 * @author Rody Kersten
 *
 */
public class TestingUnit {
	private int varCount;
	private List<JCExpression> typeExpressions;
	private String source;
	private Object classInstance;
	private Method method;
	
	/** Constructor.
	 * 
	 * @param className The name of the class in which the method to be tested resides
	 * @param methodName The name of the method to test
	 * @param string 
	 * @param argTypes The types of the method's parameters
	 * @param source The source of the class
	 * @throws CompilationException 
	 * @throws Exception
	 */
    public TestingUnit(String packageName, String className, String methodName, VarInfo variables,
            String source, String[] classPath) throws ClassNotFoundException,
            InstantiationException, IllegalAccessException,
            NoSuchMethodException, IOException, CompilationException {
        this.source = source;
        this.varCount = variables.size();
        this.typeExpressions = variables.getTypeExpressions();
        Class<?> compiledClass = compile(packageName, className, classPath);
        classInstance = compiledClass.newInstance(); // Instantiate the class

        // get the specified method
        method = compiledClass.getMethod(methodName, variables.getTypes()
                .toArray(new Class<?>[0]));
    }

    private Class<?> compile(String packageName, String className, String[] classPath)
            throws IOException, ClassNotFoundException, CompilationException {
        File tempBin = File.createTempFile("ResAna" + className, "");
        tempBin.delete();
        try {
            tempBin.mkdir();
            File sourceFile = new File(tempBin, className + ".java");

            FileWriter writer = new FileWriter(sourceFile);
            writer.append(source);
            writer.close();
            
            String joinedClassPath = "";
            URL[] classPathURLs = new URL[classPath.length + 1];
            for (int i = 0; i < classPath.length; i++) {
                joinedClassPath += classPath[i] + File.pathSeparator;
                classPathURLs[i + 1] = new File(classPath[i]).toURI().toURL();
            }
            // Add the classpath with the annotated code as first one
            classPathURLs[0] = tempBin.toURI().toURL();
            
            String[] args = new String[] { "-nowarn", "-d", tempBin.getAbsolutePath(),
                    "-cp", joinedClassPath, sourceFile.getAbsolutePath() };
            StringWriter errorBuffer = new StringWriter();
            Main.compile(args,new PrintWriter(errorBuffer));
            String errors = errorBuffer.toString();
            if(errors.length() > 0){
            	throw new CompilationException(errors);
            }
            ClassLoader classPathLoader = new URLClassLoader(classPathURLs);
            
            // Load compiled class
            packageName = packageName.trim();
            if (!packageName.isEmpty()) {
                packageName += ".";
            }
            Class<?> compiledClass = classPathLoader.loadClass(packageName + className);
            return compiledClass;
        } finally {
            deleteDir(tempBin);
        }
    }

    public static boolean deleteDir(File dir) {
        if (dir.isDirectory()) {
            for (File child : dir.listFiles()) {
                boolean success = deleteDir(child);
                if (!success) {
                    return false;
                }
            }
        }
        return dir.delete();
    }

	/** Conducts the tests.
	 * 
	 * @param testnodes The nodes on which to test the method
	 * @return An array containing the (integer) results
	 * @throws IllegalArgumentException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IndefiniteLoopException
	 */
	public int[] doTests(List<int[]> testnodes) throws IllegalArgumentException, InvocationTargetException, IllegalAccessException, IndefiniteLoopException {
		//Get result for each node
		int[] results = new int[testnodes.size()];
		int num;
		for (int i = 0; i < testnodes.size(); i++) {
			Object[] varArgs = new Object[varCount];
			for (int k = 0; k < varCount; k++) {
				if (typeExpressions.get(k) instanceof JCArrayTypeTree) {
					num = (int)testnodes.get(i)[k];
					try {
						Class<?> type = TypeScanner.toClass(((JCExpression)((JCArrayTypeTree)typeExpressions.get(k)).elemtype));
						varArgs[k] = Array.newInstance(type, num);
						// Fill the array
						for (int j = 0; j < num; j++) {
							Array.set(varArgs[k], j, num); // Try setting the value to be integer
						}
					} catch (UnsupportedTypeException e) { 
						// Do nothing
					}
				} else {
					varArgs[k] = new Integer((int)testnodes.get(i)[k]);
				}					
			}
			results[i] = executeMethod(varArgs);
		}
		return results;
	}
	
	@SuppressWarnings("deprecation")
	private Integer executeMethod (Object[] varArgs) throws IllegalArgumentException, InvocationTargetException, IllegalAccessException, IndefiniteLoopException {
		if (varArgs.length != varCount)
			throw new IllegalArgumentException("Incorrect number of values");

		Object[] args = new Object[varCount];
		
		for (int i = 0; i < varCount; i++) {
				args[i] = varArgs[i];
		}
		
		//Start method in new thread
		MethodInvocationThread mt = new MethodInvocationThread(method,classInstance,args);
		mt.start();
     
		//poll each 'interval' if the thread has finished, increase interval
		int interval = 2;
		for (int millisec = 0; millisec < 16384; millisec += interval) {
			try {
				Thread.sleep(interval);
				interval *= 2;
			} catch (InterruptedException e){
			}
			if (mt.returnVal != -1) {
				break;
			}
		}
		
		//If a return value has not been found after 10 seconds,
		// assume that loop is indefinite
		if (mt.returnVal == -1) {
			
			//kill thread
			//The stop method is deprecated, but i think there is no other
			// way to do this
			mt.stop();
			
			//throw exception
			throw new IndefiniteLoopException(args);
		}
		
		return mt.returnVal;
	}
	
	/** Gets the source for this testing unit.
	 * 
	 * @return The source code for this testing unit.
	 */
	public String getSource () {
		return source;
	}
}
