package loopbounds;

import loopbounds.parsetree.*;

import java.util.List;
import java.util.ArrayList;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCExpression;
import com.sun.tools.javac.tree.JCTree.JCStatement;

/**
 * Class to store variables and link them to their types and placeholders.
 *
 * @author Rody Kersten
 * @author Joel Haasnoot
 * @author Nick Gubbels
 *
 */
public class VarInfo {

	private List<String> fullNames;
	private List<String> placeholders;
	private List<Class<?>> types;
	private List<JCExpression> typeExpressions;
	
	private List<String> refactor;
	
	/** Constructor. Builds a VarInfo object from a list of variable names (e.g. 'next') and
	 * a list of full names (e.g. 'this.list.next').
	 * 
	 * @param variables List of variable names
	 * @param fullVariables List of full variable names
	 */
	public VarInfo(List<String> fullVariables) {
		this.fullNames = fullVariables;
		placeholders = new ArrayList<String>();
		types =  new ArrayList<Class<?>>();
		typeExpressions = new ArrayList<JCExpression>();
		refactor = new ArrayList<String>();
		
		// fill the lists
		for (String n : fullVariables) {
			placeholders.add(n.toString());
			types.add(null);
			typeExpressions.add(null);
		}
	}
	
	/** Gets the size of the lists of variables, i.e. the number of variables */
	public int size() {
		return fullNames.size();
	}
	
	/** Checks if the given variable is a field or an array.
	 * 
	 * @param i The index of the variable to check.
	 * @return True if the variable with index 'i' is a field or an array.
	 */
	public boolean isFieldOrArray(int i) {
		return isArray(i) || isFieldAccess(i);
	}
	
	/** Checks if the type of the variable at index i is array.
	 * 
	 * @param i The index of the variable
	 * @return True if i is an array type
	 */
	public boolean isArray(int i) {
		return types.get(i).isArray();
	}
	
	/** Checks if the variable at index i is an array-access (example arr[2]).
	 * 
	 * @param i The index of the variable
	 * @return True if i is an array access
	 */
	public boolean isArrayAccess(int i) {
		if (!types.get(i).isArray())
			return false;
		return fullNames.get(i).endsWith("]");
	}
	
	/** Checks if the variable i is a field access.
	 * 
	 * @param i The index of the variable
	 * @return True if i is a field access
	 */
	public boolean isFieldAccess(int i) {
		return fullNames.get(i).contains(".");
	}
		
	/** Get the list of full variable names.
	 * 
	 * @return A list of full variable names.
	 */
	public List<String> getFullNames() {
		return fullNames;
	}
	
	/** Get the full variable name.
	 * 
	 * @return The full variable name.
	 */
	public String getFullName(int i) {
		return fullNames.get(i);
	}
	
	/** Get the list of placeholders. An example of a placeholder is
	 * 'this_a' for 'this.a'.
	 * 
	 * @return The placeholder names
	 */
	public List<String> getPlaceholders() {
		return placeholders;
	}
	
	/** Get a placeholder. An example of a placeholder is
	 * 'this_a' for 'this.a'.
	 * 
	 * @param i The index of the placeholder to get
	 * @return The placeholder
	 */
	public String getPlaceHolder(int i) {
		return placeholders.get(i);
	}
	
	/** Set a placeholder. An example of a placeholder is
	 * 'this_a' for 'this.a'.
	 * 
	 * @param i The index of the placeholder to set
	 * @param var What to set it to
	 * @return The placeholder that has been set
	 */
	public String setPlaceHolder(int i, String var) {
		return placeholders.set(i, var);
	}
	
	/** Get the types of the variables
	 * 
	 * @return A list of the variable's types
	 */
	public List<Class<?>> getTypes() {
		return types;
	}
	
	/** Get the type of a variable.
	 * 
	 * @param i The index of the variable
	 * @return The type of the variable
	 */
	public Class<?> getType(int i) {
		return types.get(i);
	}
	
	/** Set the type of a variable.
	 * 
	 * @param i The index of the variable
	 * @param type What to set it to
	 * @return The type that the variable was set to
	 */
	public Class<?> setType(int i, Class<?> type) {
		return types.set(i, type);
	}
	
	/** Get the variable types in JCExpression form
	 * 
	 * @return A list of JCExpressions representing the types.
	 */
	public List<JCExpression> getTypeExpressions() {
		return typeExpressions;
	}
	
	/** Get a variable type in JCExpression form
	 * 
	 * @param i The index of the variable for which to get the type
	 * @return A JCExpression representing the type of the variable.
	 */
	public JCExpression getTypeExpression(int i) {
		return typeExpressions.get(i);
	}
	
	/** Set a variable type in JCExpression form
	 * 
	 * @param i The index of the variable for which to set the type
	 * @param type A JCExpression representing the type of the variable
	 * @return True if it was succesful and false (i.e. there was refactoring) otherwise
	 */
	public boolean setType(int i, JCExpression type) throws UnsupportedTypeException {
		try{
			typeExpressions.set(i, type);
			types.set(i, TypeScanner.toClass(type));
			return true;
		}catch(UnsupportedTypeException e){ //Try to solve with refactoring
			refactor(i);
			return false;
		}
	}
	
	/** Check if a variable has been replaced by a placeholder (e.g. this.a -> this_a).
	 * 
	 * @param i The index of the variable
	 * @return True if the variable has been replaced by a placeholder
	 */
	public boolean adapted(int i) {
		return !fullNames.get(i).equals(placeholders.get(i));
	}
	
	/** Load the types of the variables into the data-structure.
	 * 
	 * @param code The parse tree under analysis
	 * @param classIndex The class to load types from
	 * @param methIndex The method to load types from
	 * @throws UnsupportedTypeException
	 * @throws UndeclaredVariableException
	 */
	public void loadTypes(ParseTreeWrapper code, int classIndex, int methIndex)
		throws UnsupportedTypeException, UndeclaredVariableException {
		
		if (LoopBoundsAnalyzer.DEBUG)
			System.out.println("[VarInfo] Trying to load types for: " + fullNames);
		
		ArrayList<CodeClass> codeclasses = ClassTypeScanner.getTypes(code.getInputTree());

		for (int i = 0; i < size(); i++) {
			try {

				if (!setType(
						i,
						TypeScanner.getExpression(getFullName(i),
								code.getMethod(classIndex, methIndex),
								code.getInputTree(), codeclasses))) {
					i--;
				}
			} catch (UndeclaredVariableException e) { //Try refactoring
				refactor(i);
				i--;
			}
		}
		
		if (LoopBoundsAnalyzer.DEBUG)
			System.out.println("[VarInfo] Found: " + typeExpressions);
	}
	
	/** Find a certain full name.
	 * 
	 * @param n The name to look for
	 * @return The index of the name if found, else -1
	 */
	public int findFullName(String n) {
		for (int i = 0; i < size(); i++)
			if (fullNames.get(i).equals(n))
				return i;
		return -1;
	}
	
	public List<String> getRefactoredObjects(){
		return refactor;
	}

	/** Set variable to be refactored
	 * 
	 * @param i The index of the variable to be refactored
	 */
	public void refactor(int i){
		refactor.add(fullNames.get(i));
		remove(i);
	}
	
	/** Remove a variable.
	 * 
	 * @param i The index of the variable to remove
	 */
	public void remove(int i) {
		fullNames.remove(i);
		placeholders.remove(i);
		types.remove(i);
		typeExpressions.remove(i);
	}
	
	/** Rename all variables of field and array types to a placeholder of the
	 * element type. Example: replace 'this.num' by 'num_0' where the second has
	 * the same type as the first (integer). Placeholders are stored in the 'placeholders'
	 * list.
	 * 
	 * @param code The parse tree under analysis
	 * @param classIndex The index of the class to rename variables in
	 * @param methIndex The index of the method to rename variables in
	 * @param loopIndex The index of the loop to rename variables in
	 * @throws UnsupportedTypeException
	 * @throws UndeclaredVariableException
	 */
	public void rename(ParseTreeWrapper code, int classIndex, int methIndex, int loopIndex)
		throws UnsupportedTypeException, UndeclaredVariableException {
	
		loadTypes(code, classIndex, methIndex);
		JCStatement loopCopy = code.getLoop(classIndex, methIndex, loopIndex);
	
		for (int i = 0; i < size(); i++) {
			if (isFieldOrArray(i)) {
				
				JCExpression type = getTypeExpression(i);
				String fullName = getFullName(i);

				if (type instanceof JCTree.JCArrayTypeTree && fullName.endsWith("]")) {

                    // Array access -> replace the whole array index part with the current variable index
				    // This might result in semantically different behavior of the code (with for example a[i + 1]
				    // But otherwise the generated code wont be valid syntacticly
                    String newName = fullName.substring(0,fullName.indexOf('[')) + "_" + i;
					
					//Check if newName exists, if so, append an index
					if (VariableScanner.containsVar(loopCopy, newName)) {
						int j = 0;
						while (VariableScanner.containsVar(loopCopy, newName+"_"+j)) j++;
						newName = newName+"_"+j;
					}
					fullNames.set(i,fullName);
					typeExpressions.set(i,((JCTree.JCArrayTypeTree) type).elemtype);
					types.set(i,TypeScanner.toClass(((JCTree.JCArrayTypeTree) type).elemtype));
					placeholders.set(i, newName);

				} else {
					
					//Field -> replace by var with . replaced by _
					String newName = fullName.replace('.', '_');

					//Check if newName exists, if so, append an index
					if (VariableScanner.containsVar(loopCopy, newName)) {
						int j = 0;
						while (VariableScanner.containsVar(loopCopy, newName+"_"+j)) j++;
						newName = newName+"_"+j;
					}
					setPlaceHolder(i, newName);
					
					// Exceptional case, for array.length
					if (type instanceof JCTree.JCArrayTypeTree && fullName.endsWith(".length")) {
						typeExpressions.set(i,((JCTree.JCArrayTypeTree) type).elemtype);
						types.set(i,TypeScanner.toClass(((JCTree.JCArrayTypeTree) type).elemtype));
					}
				}
			}
		}
		
		if (LoopBoundsAnalyzer.DEBUG) {
			System.out.println("Final variables: " + fullNames);
			System.out.println("Placeholders: " + placeholders);
		}
	}
}
