package loopbounds.parsetree;

import loopbounds.LoopBoundsAnalyzer;

import java.util.List;
import java.util.ArrayList;

import org.jmlspecs.openjml.JmlTree;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.*;
import com.sun.tools.javac.util.Context;

/**
 * Class which implements some transformations on boolean JCExpressions.
 * 
 * @author Rody Kersten
 *
 */
public class ConditionTransformer {
	
	JmlTree.Maker M;

	/** Default constructor. Needs the context from OpenJDK. */
	public ConditionTransformer(Context context) {
		this.M = JmlTree.Maker.instance(context);
	}
	
	/** Transforms a logical expression into DNF.
	 * 
	 * Example, when the input is 'b1 AND (b2 OR b3)', the output will
	 * be '(b1 AND b2) OR (b1 AND b3)'.
	 * 
	 * @param exp A logical JCExpression over elementary propositions.
	 * @return The expression in Disjunctive Normal Form.
	 */
	public JCExpression DNF(JCExpression exp) {
		JCExpression distr = distributeNegationsAux(exp, false);
		return transformAux(distr);
	}
	
	/** Walks through the parse tree and distributes negations.
	 * In other words, moves all negations to the elementary propositions.
	 * 
	 * Also removes parentheses for logical functions. Not for elementary propositions!
	 * 
	 * @param exp A logical JCExpression over elementary propositions.
	 * @return The expression with all negations moved to the elementary propositions, i.e. the leafs of the tree, and unnecessary parentheses removed.
	 **/
	public JCExpression distributeNegations(JCExpression exp) {
		return distributeNegationsAux(exp,false);
	}
	
	/** Gets the number of ORs at the top of the syntax tree.
	 * Used for retrieving the number of pieces of the piecewise polynomial.
	 * 
	 * Example, if the input is 'b1 OR b2', the output is 1.
	 * 
	 * @param exp A logical JCExpression over elementary propositions.
	 * @return The number of ORs at the top of the tree.
	 */
	public int getNumORs(JCExpression exp) {
		if ((exp instanceof JCBinary) && (exp.getTag() == JCTree.OR)) {
			JCBinary or = (JCBinary) exp;
			return 1 + getNumORs(or.lhs) + getNumORs(or.rhs);
		}
		return 0;
	}
	
	/** Splits the syntax tree up into separate pieces, based on ORs.
	 * 
	 * Example, when the input is 'b1 OR b2', the output is a list
	 * containing b1 \ b2, b2 \ b1 and b1 AND b2.
	 * 
	 * @param exp A logical JCExpression over elementary propositions.
	 * @return A list of the logical expressions separated by ORs in 'exp'.
	 */
	public List<JCExpression> splitOR(JCExpression exp) {
		List<JCExpression> todo = splitORNaive(exp);
		List<List<JCExpression>> subs = getAllSubsets(todo);
		List<JCExpression> result = new ArrayList<JCExpression>();
		
		for (List<JCExpression> subset : subs) {
			result.add(combineANDs(subset));
		}
		
		if (LoopBoundsAnalyzer.DEBUG)
			System.out.println("Disjunction-splitting done, found the following pieces: " + result);
		
		return result;
	}
	
	private <T extends Object> List<List<T>> getAllSubsets(List<T> lst) {
		List<List<T>> result = new ArrayList<List<T>>();
		
		if (!lst.isEmpty()) {
			
			//pop the head of the list
			T head = lst.remove(0);
			
			//add the element to the result
			List<T> explist = new ArrayList<T>();
			explist.add(head);
			result.add(explist);
			
			//recursive call
			List<List<T>> subs = getAllSubsets(lst);
			
			//add all subsets of the tail to the result
			result.addAll(subs);
			
			//combine head with all subsets of the tail
			for (List<T> subset : subs) {
				List<T> copy = new ArrayList<T>();
				for (T t : subset)
					copy.add(t);
				copy.add(head);
				result.add(copy);
			}
		}
		
		return result;
	}
	
	private JCExpression combineANDs(List<JCExpression> exps) {
		if (exps.size() == 1) {
			return exps.get(0);
		} else {
			return M.Binary(
				JCTree.AND,
				exps.remove(0),
				combineANDs(exps)
			);
		}
	}
	
	/** Splits the syntax tree up into separate pieces, based on ORs.
	 * 
	 * Example, when the input is 'b1 OR b2', the output is a list containing b1 and b2.
	 * 
	 * @param exp A logical JCExpression over elementary propositions.
	 * @return A list of the logical expressions separated by ORs in 'exp'.
	 */
	public List<JCExpression> splitORNaive(JCExpression exp) {
		if (exp instanceof JCParens) {
			JCParens parens = (JCParens) exp;
			return splitORNaive(parens.expr);
		}
		
		List<JCExpression> l = new ArrayList<JCExpression>();
		if ((exp instanceof JCBinary) && (exp.getTag() == JCTree.OR)) {
			JCBinary or = (JCBinary) exp;
			l.addAll(splitORNaive(or.lhs));
			l.addAll(splitORNaive(or.rhs));
		} else {
			l.add(exp);
		}
		return l;
	}
	
	/** Splits the syntax tree up into separate pieces, based on ANDs.
	 * 
	 * Example, when the input is 'b1 AND b2', the output is a list containing b1 and b2.
	 * 
	 * @param exp A logical JCExpression over elementary propositions.
	 * @return A list of the logical expressions separated by ANDs in 'exp'.
	 */
	public List<JCExpression> splitAND(JCExpression exp) {
		if (exp instanceof JCParens) {
			JCParens parens = (JCParens) exp;
			return splitAND(parens.expr);
		}
		
		List<JCExpression> l = new ArrayList<JCExpression>();
		if ((exp instanceof JCBinary) && (exp.getTag() == JCTree.AND)) {
			JCBinary and = (JCBinary) exp;
			l.addAll(splitAND(and.lhs));
			l.addAll(splitAND(and.rhs));
		} else {
			l.add(exp);
		}
		return l;
	}
	
	/** Removes an optional negation around a linear (in)equality.
	 * Example: transforms !(a > 4) into '(a <= 4)'.
	 *
	 * @param exp A logical JCExpression over elementary propositions.
	 * @return The expression with negations removed.
	 */
	public JCExpression removeNeg(JCExpression exp) {
		if (exp instanceof JCParens) {
			JCParens parens = (JCParens) exp;
			return removeNeg(parens.expr);
		} else if ((exp instanceof JCUnary) && (exp.getTag() == JCTree.NOT)) {
			JCUnary not = (JCUnary) exp;
			try { 
				return negateInequality(not.arg);
			} catch (Exception e) {
				System.out.println("Unable to remove negation around expression '" + not.arg + "', because it is not an (in)equality.\n");
				return not;
			}
		} else {
			return exp;
		}
	}
	
	/** Negates a linear (in)equality. Auxiliary to removeNeg.
	 * Example: transforms 'a > 4' into 'a <= 4'.
	 */
	private JCExpression negateInequality(JCExpression exp) throws Exception {
		if (exp instanceof JCParens) {
			JCParens parens = (JCParens) exp;
			return M.Parens(negateInequality(parens.expr));
		}
		
		if (exp instanceof JCBinary) {
			JCBinary bin = (JCBinary) exp;
			if (bin.getTag() == JCTree.EQ)
				return M.Binary(JCTree.NE, bin.lhs, bin.rhs);
			else if (bin.getTag() == JCTree.NE)
				return M.Binary(JCTree.EQ, bin.lhs, bin.rhs);
			else if (bin.getTag() == JCTree.GE)
				return M.Binary(JCTree.LT, bin.lhs, bin.rhs);
			else if (bin.getTag() == JCTree.LT)
				return M.Binary(JCTree.GE, bin.lhs, bin.rhs);
			else if (bin.getTag() == JCTree.GT)
				return M.Binary(JCTree.LE, bin.lhs, bin.rhs);
			else if (bin.getTag() == JCTree.LE)
				return M.Binary(JCTree.GT, bin.lhs, bin.rhs);
		}
		throw new Exception();
	}
	
	private JCExpression distributeNegationsAux(JCExpression exp, boolean inv) {
		
		if (exp instanceof JCParens) {
			JCParens parens = (JCParens) exp;
			
			//return M.Parens(distributeNegationsAux(parens.expr,M, inv));
			return distributeNegationsAux(parens.expr, inv);
		} else if ((exp instanceof JCUnary) && (exp.getTag() == JCTree.NOT)) {
			JCUnary not = (JCUnary) exp;
			return distributeNegationsAux(not.arg, !inv);
		} else if ((exp instanceof JCBinary) && (exp.getTag() == JCTree.AND)) {
			JCBinary and = (JCBinary) exp;
			if (inv)
				return M.Binary(
						JCTree.OR,
						distributeNegationsAux(M.Unary(JCTree.NOT, and.lhs), !inv),
						distributeNegationsAux(M.Unary(JCTree.NOT, and.rhs), !inv)
				);
			return M.Binary(
					JCTree.AND,
					distributeNegationsAux(and.lhs, inv),
					distributeNegationsAux(and.rhs, inv)
			);
		} else if ((exp instanceof JCBinary) && (exp.getTag() == JCTree.OR)) {
			JCBinary or = (JCBinary) exp;
			if (inv)
				return M.Binary(
						JCTree.AND,
						distributeNegationsAux(M.Unary(JCTree.NOT, or.lhs), !inv),
						distributeNegationsAux(M.Unary(JCTree.NOT, or.rhs), !inv)
				);
			return M.Binary(
					JCTree.OR,
					distributeNegationsAux(or.lhs, inv),
					distributeNegationsAux(or.rhs, inv)
			);
		} else {
			//Elementary proposition
			
			//add parentheses again
			exp = M.Parens(exp);
			
			if (inv)
				return M.Unary(JCTree.NOT, exp);
			
			return exp;
		}

	}

	/** Combines all expressions other than ORs from a and b into ANDs */
	private JCExpression pairs(JCExpression a, JCExpression b) {
		if ((a instanceof JCBinary) && (a.getTag() == JCTree.OR)) {
			JCBinary or = (JCBinary) a;		
			return M.Binary(
					JCTree.OR,
					pairs(or.lhs, b),
					pairs(or.rhs, b)
			);
		} else if ((b instanceof JCBinary) && (b.getTag() == JCTree.OR)) {
			JCBinary or = (JCBinary) b;
			return M.Binary(
					JCTree.OR,
					pairs(a, or.lhs),
					pairs(a, or.rhs)
			);
		} else {
			//Done!
			return M.Parens(M.Binary(JCTree.AND, a, b));
		}
	}

	private JCExpression transformAux(JCExpression exp) {
		
		if ((exp instanceof JCBinary) && (exp.getTag() == JCTree.AND)) {
			JCBinary and = (JCBinary) exp;
			JCExpression lhs = transformAux(and.lhs);
			JCExpression rhs = transformAux(and.rhs);
			
			if ( ((lhs instanceof JCBinary) && (lhs.getTag() == JCTree.OR))
				|| ( ((rhs instanceof JCBinary) && (rhs.getTag() == JCTree.OR))) ) {
				//Action is needed!
				return pairs(lhs, rhs);
			}

			return M.Binary(JCTree.AND,	lhs, rhs);
		} else if ((exp instanceof JCBinary) && (exp.getTag() == JCTree.OR)) {
			JCBinary or = (JCBinary) exp;
			return M.Binary(
					JCTree.OR,
					transformAux(or.lhs),
					transformAux(or.rhs)
			);
		} else {
			//Elementary proposition
			return exp;
		}
	}
}
