package loopbounds.parsetree;

import loopbounds.VarInfo;

import org.jmlspecs.openjml.JmlTree;

import com.sun.tools.javac.code.Type;
import com.sun.tools.javac.code.TypeTags;
import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCMethodDecl;
import com.sun.tools.javac.tree.JCTree.JCStatement;
import com.sun.tools.javac.tree.JCTree.JCTypeParameter;
import com.sun.tools.javac.tree.JCTree.JCVariableDecl;
import com.sun.tools.javac.util.ListBuffer;
import com.sun.tools.javac.util.Name;
import com.sun.tools.javac.util.Names;

/**
 * Class for creating an abstract syntax tree of a new method. The class is
 * intended to be used through the static method 'createMethod'.
 * 
 * @author Rody Kersten
 * 
 */
public class MethodCreator {

	private JCStatement loop;
	private JmlTree.Maker M;
	private ParseTreeWrapper code;
	private int classIndex, methIndex;
	private VarInfo variables;
	
	/** Default constructor */
	private MethodCreator(JCStatement loop,	ParseTreeWrapper code, int classIndex, int methIndex, VarInfo variables) {
		this.loop = loop;
		this.variables = variables;
		this.code = code;
		this.classIndex = classIndex;
		this.methIndex = methIndex;
		this.M = JmlTree.Maker.instance(code.context());
	}

	/**
	 * Creates a method containing only the given loop, for testing and
	 * verification. Some information about the rest of the class should also be
	 * known, to find the types of variables and to make sure that a method name
	 * is chosen that does not exist yet.
	 * 
	 * @param loop
	 *            The loop to create the method around
	 * @param code
	 *            The parse tree in which to add the method
	 * @param classIndex
	 * 			  The index of the class in which to add the method
	 * @param methIndex 
	 *            The index of an existing method on which the new method will be based.
	 *            Return type and thrown exceptions of this method may be used.
	 * @param variables
	 *            The variables used in the loop, i.e. the parameters to the new method.
	 * @param addCounter
	 *            Whether or not the loop should be instrumented with a counter
	 * @return The new method
	 * @throws UnsupportedTypeException
	 */
	public static JCMethodDecl createMethod(JCStatement loop,
			ParseTreeWrapper code, int classIndex, int methIndex, VarInfo variables,
			boolean addCounter) throws UnsupportedTypeException {
		MethodCreator creator = new MethodCreator(loop, code, classIndex, methIndex, variables);
		return creator.create(addCounter);
	}

	private JCMethodDecl create(boolean addCounter)
			throws UnsupportedTypeException {
		// Set parameters
		ListBuffer<JCVariableDecl> params = new ListBuffer<JCVariableDecl>();

		JCMethodDecl fromMeth = code.getMethod(classIndex, methIndex);
		Name methodName = newMethodName(fromMeth.name.toString());

		// get the undeclared variables
		for (int i = 0; i < variables.size(); i++) {
				params.append(M.VarDef(
						M.Modifiers(0),
						M.Name(variables.getPlaceHolder(i)),
						variables.getTypeExpression(i),
						null)
					);
		}

		Name countName = newVarName("count");

		ListBuffer<JCStatement> stats = new ListBuffer<JCStatement>();

		// Declare counter variable
		if (addCounter) {

			stats.append(M.VarDef(M.Modifiers(0), countName, M.Type(new Type(
					TypeTags.INT, null)), M.Literal(0)));

			// Add loop
			augmentLoop(countName, loop);
			ReturnReplacer.replace(countName, loop, code.context());
		}

		stats.append(loop);

		// Add return statement
		if (addCounter) {
			stats.append(M.Return(M.Ident(countName)));
		}
		
		JCMethodDecl newMethod = M
				.MethodDef(M.Modifiers(1),
						methodName,
						// addCounter ? M.TypeIdent(TypeTags.INT) :
						// (JCExpression) fromMeth.getReturnType(),
						addCounter ? M.TypeIdent(TypeTags.INT) : M
								.TypeIdent(TypeTags.VOID),
						com.sun.tools.javac.util.List.<JCTypeParameter> nil(),
						params.toList(), fromMeth.thrown, M.Block(0, stats
								.toList()), null);

		return newMethod;
	}

	/**
	 * Transforms the class type into a TypeTag, or throws an exception if the
	 * type is unsupported.
	 * 
	 * @param type
	 *            The class type
	 * @return The corresponding TypeTag
	 * @throws UnsupportedTypeException
	 */
	public static int getTypeTag(Class<?> type) throws UnsupportedTypeException {
		if (type.equals(int.class)) {
			return TypeTags.INT;
		}
		if (type.equals(short.class)) {
			return TypeTags.SHORT;
		}
		if (type.equals(long.class)) {
			return TypeTags.LONG;
		}
		if (type.equals(byte.class)) {
			return TypeTags.BYTE;
		}
		if (type.equals(char.class)) {
			return TypeTags.CHAR;
		}
		if (type.equals(float.class)) {
			return TypeTags.FLOAT;
		}
		if (type.equals(double.class)) {
			return TypeTags.DOUBLE;
		}
		if (type.equals(int[].class)) {
			return TypeTags.ARRAY;
		}
		throw new UnsupportedTypeException(type.getName());
	}

	 //Creates a new name for the method, taking care the it does not already exist
	 private Name newMethodName (String basename) {
		 int i;
		 for (i = 0; code.methodExists(basename+"_"+i, classIndex); i++);
		 return Names.instance(code.context()).fromString(basename+"_"+i);
	 }

	// Creates a new variable name, taking care the it does not already exist
	private Name newVarName(String prefix) {
		Name newName;
		int i = 0;
		do {
			newName = Names.instance(code.context()).fromString(prefix + "_" + i);
			i++;
		} while (variables.getFullNames().contains(newName.toString())
				|| VariableScanner.containsVar(loop, newName));

		return newName;
	}

	/**
	 * Augments the given loop with a counter for the number of iterations of
	 * the loop.
	 * 
	 * @param countName
	 * @param tree
	 */
	private void augmentLoop(Name countName, JCStatement loop) {
		ListBuffer<JCStatement> stats = new ListBuffer<JCStatement>();

		JCTree.JCIdent cname = M.Ident(countName);

		JCTree.JCUnary inc = M.Unary(JCTree.POSTINC, cname);
		stats.append(M.Exec(inc));

		if (loop instanceof JmlTree.JmlWhileLoop) {
			JmlTree.JmlWhileLoop whileloop = (JmlTree.JmlWhileLoop) loop;
			stats.append(whileloop.body);
			whileloop.body = M.Block(0, stats.toList());
		} else if (loop instanceof JmlTree.JmlDoWhileLoop) {
			JmlTree.JmlDoWhileLoop doloop = (JmlTree.JmlDoWhileLoop) loop;
			stats.append(doloop.body);
			doloop.body = M.Block(0, stats.toList());
		} else {
			JmlTree.JmlForLoop forloop = (JmlTree.JmlForLoop) loop;
			stats.append(forloop.body);
			forloop.body = M.Block(0, stats.toList());
		} 

		// Add count < n to the condition, to stop indefinite loops
		// TODO: somehow this doesn't work, now uses threading and Thread.stop()
		// JCTree.JCExpression stopper = M.Parens(M.Binary(JCTree.LT, cname,
		// M.Literal(2)));
		// loop.cond = M.Parens(M.Binary(JCTree.AND, loop.cond, stopper));
	}
}
