package loopbounds.parsetree;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;

import org.jmlspecs.openjml.API;
import org.jmlspecs.openjml.JmlTree;
import org.jmlspecs.openjml.JmlTree.JmlMethodDecl;
import org.jmlspecs.openjml.JmlTree.JmlSpecificationCase;

import com.sun.tools.javac.util.Context;
import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.*;

public class ParseTreeWrapper {
	private API api;
	private String source = null;
	private StringWriter diagnostic;
	private JmlTree.JmlCompilationUnit parsetree = null;
	private List<JmlTree.JmlClassDecl> classes = new ArrayList<JmlTree.JmlClassDecl>();
	private List<List<JmlTree.JmlMethodDecl>> methods = new ArrayList<List<JmlTree.JmlMethodDecl>>();
	private List<List<List<JCStatement>>> loops = new ArrayList<List<List<JCStatement>>>();
	
	
	/** Default constructor. 
	 * 
	 * @param diagnostic StringWriter to write errors to
	 * @throws java.io.IOException
	 */
	public ParseTreeWrapper(StringWriter diagnostic) throws java.io.IOException {
		this.diagnostic = diagnostic;
		api = new API(new PrintWriter(diagnostic), null);
	}

	/**
	 * Parse source code.
	 * 
	 * @param code The source to be parsed
	 * */
	public void parseSource(String code) throws Exception {
		source = code;
		parsetree = api.parseString("input", source);
		classes.clear();
		methods.clear();
		loops.clear();
		classes = ClassScanner.getClasses(parsetree);
		for (int i = 0; i < classes.size(); i++) {
			methods.add(i, MethodScanner.getMethods(classes.get(i)));
			loops.add(i, new ArrayList<List<JCStatement>>());
			for (int j = 0; j < methods.get(i).size(); j++) {
				loops.get(i).add(j, LoopScanner.getLoops(methods.get(i).get(j)));
			}
			assert(methods.get(i).size() == loops.get(i).size());
		}
		assert(classes.size() == methods.size());
	}
	
	
	/** Loads a new ParseTreeWrapper by parsing the source code again.
	 * Changes made in the parse tree are ignored.
	 */
	public ParseTreeWrapper freshcopy() throws Exception {
		ParseTreeWrapper ret = new ParseTreeWrapper(diagnostic);
		ret.parseSource(source);
		return ret;
	}
	
	/** Loads a copy of this ParseTreeWrapper.
	 * 
	 * Pretty prints the code and parses this again.
	 */
	public ParseTreeWrapper copy() throws Exception {
		ParseTreeWrapper ret = new ParseTreeWrapper(diagnostic);
		ret.parseSource(this.prettyPrint());
		return ret;
	}
	
	/** Get the entire tree */
	public JmlTree.JmlCompilationUnit getInputTree() {
		return parsetree;
	}
	
	/** Get the source code */
	public String getSource() {
		return source;
	}
	
	/** Get the parse trees of the classes.
	 * 
	 * @return The parse trees of the classes.
	 */
	public List<JmlTree.JmlClassDecl> getClasses() {
		return classes;
	}
	
	/** Get the number of classes.
	 * 
	 * @return The number of classes.
	 */
	public int getNumClasses() {
		return classes.size();
	}
	
	/** Get the parse tree of a class by index.
	 * 
	 * @param classIndex The index of the class to get the parse tree from.
	 * @return The parse tree of class classIndex.
	 */
	public JmlTree.JmlClassDecl getClass(int classIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()))
			throw new ParseTreeIndexOutOfBoundsException("Illegal class selected");
		return classes.get(classIndex);
	}
	
	/**
	 * Get the names of the classes in the parsed code. Useful because a class,
	 * method and loop may need to be selected for analysis by a user interface.
	 * 
	 * @return A List of Strings containing the names of the classes in the
	 *         parsed code.
	 */
	public List<String> getClassNames() {
		List<String> classNames = new ArrayList<String>();

		for (int i = 0; i < classes.size(); i++)
			classNames.add(i, classes.get(i).name.toString());

		return classNames;
	}
	
	/** Get the name of a class by index.
	 * 
	 * @param classIndex The index of the class to get the name from.
	 * @return A String representing the name of class classIndex.
	 */
	public String getClassName(int classIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()))
			throw new ParseTreeIndexOutOfBoundsException("Illegal class selected");
		return classes.get(classIndex).name.toString();
	}
	
	
	/** Get the parse trees of the methods.
	 * 
	 * @return The parse trees of the classes.
	 */
	public List<JmlTree.JmlMethodDecl> getMethods(int classIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()))
			throw new ParseTreeIndexOutOfBoundsException("Illegal class selected");
		return methods.get(classIndex);
	}
	
	/** Get the number of methods in a class.
	 * 
	 * @return The number of methods in the class
	 */
	public int getNumMethods(int classIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()))
			throw new ParseTreeIndexOutOfBoundsException("Illegal class selected");
		return methods.get(classIndex).size();
	}
	
	/** Get the parse tree of a method by index.
	 * 
	 * @param classIndex The index of the class to get the parse tree from.
	 * @param methIndex The index of the method to get the parse tree from.
	 * @return The parse tree of class classIndex.
	 */
	public JmlTree.JmlMethodDecl getMethod(int classIndex, int methIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()
				&& methIndex >= 0 && methIndex < methods.get(classIndex).size()))
				throw new ParseTreeIndexOutOfBoundsException("Illegal class or method selected");
		return methods.get(classIndex).get(methIndex);
	}
	
	/**
	 * Get the names of the methods in the parsed code. Useful because a class,
	 * method and loop may need to be selected for analysis by a user interface.
	 * 
	 * @return A List of Strings containing the names of the methods in the
	 *         parsed code.
	 */
	public List<String> getMethodNames(int classIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()))
			throw new ParseTreeIndexOutOfBoundsException("Illegal class selected");
		
		List<String> methodNames = new ArrayList<String>();
		for (int i = 0; i < methods.get(classIndex).size(); i++)
			methodNames.add(i, methods.get(classIndex).get(i).name.toString());
		return methodNames;
	}
	
	/** Get the name of a method by index.
	 * 
	 * @param classIndex The index of the class to get the name from.
	 * @param methIndex The index of the method to get the name from.
	 * @return A String representing the name of the method.
	 */
	public String getMethodName(int classIndex, int methIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()
				&& methIndex >= 0 && methIndex < methods.get(classIndex).size()))
				throw new ParseTreeIndexOutOfBoundsException("Illegal class or method selected");
		return methods.get(classIndex).get(methIndex).name.toString();
	}

	
	
	/** Get the parse trees of the loops in a specified method in a specified class.
	 * 
	 * @param classIndex The index of the class to fetch the loops from
	 * @param methIndex The index of the method to fetch the loops from
	 * @return The parse trees of the loops in a list.
	 */
	public List<JCStatement> getLoops(int classIndex, int methIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()
				&& methIndex >= 0 && methIndex < methods.get(classIndex).size()))
				throw new ParseTreeIndexOutOfBoundsException("Illegal class or method selected");
		return loops.get(classIndex).get(methIndex);
	}
	
	/** Get the number of loops in a specified method in a specified class.
	 * 
	 * @param classIndex The index of the class to fetch the loops from
	 * @param methIndex The index of the method to fetch the loops from
	 * @return The number of loops
	 */
	public int getNumLoops(int classIndex, int methIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()
				&& methIndex >= 0 && methIndex < methods.get(classIndex).size()))
				throw new ParseTreeIndexOutOfBoundsException("Illegal class or method selected");
		return loops.get(classIndex).get(methIndex).size();
	}
	
	/** Get the parse tree of a loop in a specified method in a specified class.
	 * 
	 * @param classIndex The index of the class to fetch the loop from
	 * @param methIndex The index of the method to fetch the loop from
	 * @param loopIndex The index of the loop to fetch
	 * @return The parse tree of the loop
	 */
	public JCStatement getLoop(int classIndex, int methIndex, int loopIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()
				&& methIndex >= 0 && methIndex < methods.get(classIndex).size()
				&& loopIndex >= 0 && loopIndex < loops.get(classIndex).get(methIndex).size()))
				throw new ParseTreeIndexOutOfBoundsException("Illegal class, method or loop selected");
		return loops.get(classIndex).get(methIndex).get(loopIndex);
	}

	/**
	 * Get the source code of the loops inside a method. Useful because a class,
	 * method and loop may need to be selected for analysis by a user interface.
	 * 
	 * @param methodIndex The index of the method from which to extract the loops.
	 * @return A List of Strings containing the source code of the loops in the
	 *         method with index 'methodIndex'.
	 */
	public List<String> getLoopStrings(int classIndex, int methIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()
				&& methIndex >= 0 && methIndex < methods.get(classIndex).size()))
				throw new ParseTreeIndexOutOfBoundsException("Illegal class or method selected");
		
		List<String> loopStrings = new ArrayList<String>();
		for (int i = 0; i < loops.get(classIndex).get(methIndex).size(); i++)
			loopStrings.add(i, loops.get(classIndex).get(methIndex).get(i).toString());
		return loopStrings;
	}
	
	/**
	 * Get the source code of a loop inside a specified method inside a specified class.
	 * Useful because a class, method and loop may need to be selected for analysis
	 * by a user interface.
	 * 
	 * @param classIndex The index of the class to fetch the loop from
	 * @param methIndex The index of the method to fetch the loop from
	 * @param loopIndex The index of the loop to fetch
	 * @return A String containing the source code of the specified loop
	 */
	public String getLoopString(int classIndex, int methIndex, int loopIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()
				&& methIndex >= 0 && methIndex < methods.get(classIndex).size()
				&& loopIndex >= 0 && loopIndex < loops.get(classIndex).get(methIndex).size()))
				throw new ParseTreeIndexOutOfBoundsException("Illegal class, method or loop selected");
		return loops.get(classIndex).get(methIndex).get(loopIndex).toString();
	}
	
	/**
	 * Gets the line-number position of a class in the original source code
	 * 
	 * @param classIndex The index of the class
	 * @return Line number of the specified class
	 */
	public int getMethodPosition(int classIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()))
			throw new ParseTreeIndexOutOfBoundsException("Illegal class, method or loop selected");
		return classes.get(classIndex).pos;
	}
	
	/**
	 * Gets the line-number position of a method in the original source code
	 * 
	 * @param classIndex The index of the class
	 * @param methIndex The index of the method
	 * @return Line number of the specified method
	 */
	public int getMethodPosition(int classIndex, int methIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()
				&& methIndex >= 0 && methIndex < methods.get(classIndex).size()))
			throw new ParseTreeIndexOutOfBoundsException("Illegal class, method or loop selected");
		
		int pos = methods.get(classIndex).get(methIndex).getModifiers().pos;
		
		//if no modifiers, take return type
		if (pos < 0)
			pos = methods.get(classIndex).get(methIndex).getStartPosition();
		
		return pos;
	}
	
	/**
	 * Gets the line-number position of a loop in the original source code
	 * 
	 * @param classIndex The index of the class to fetch the loop from
	 * @param methIndex The index of the method to fetch the loop from
	 * @param loopIndex The index of the loop to fetch
	 * @return Line number of the specified loop
	 */
	public int getLoopPosition(int classIndex, int methIndex, int loopIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()
				&& methIndex >= 0 && methIndex < methods.get(classIndex).size()
				&& loopIndex >= 0 && loopIndex < loops.get(classIndex).get(methIndex).size()))
			throw new ParseTreeIndexOutOfBoundsException("Illegal class, method or loop selected");
		return loops.get(classIndex).get(methIndex).get(loopIndex).pos;
	}
	
	/** Checks if a method is JML annotated
	 * 
	 * @param classIndex The index of the class
	 * @param methIndex The index of the method
	 * @return True if the method is JML annotated
	 */
	public boolean isAnnotated(int classIndex, int methIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()
				&& methIndex >= 0 && methIndex < methods.get(classIndex).size()))
			throw new ParseTreeIndexOutOfBoundsException("Illegal class, method or loop selected");
		JmlMethodDecl meth = getMethod(classIndex,methIndex);
		return meth.cases != null && meth.cases.cases != null && !meth.cases.cases.isEmpty();
	}
	
	/**
	 * Checks if a loop is JML annotated
	 * 
	 * @param classIndex The index of the class to fetch the loop from
	 * @param methIndex The index of the method to fetch the loop from
	 * @param loopIndex The index of the loop to fetch
	 * @return True if the loop is JML annotated
	 */
	public boolean isAnnotated(int classIndex, int methIndex, int loopIndex) {
		if (!(classIndex >= 0 && classIndex < classes.size()
				&& methIndex >= 0 && methIndex < methods.get(classIndex).size()
				&& loopIndex >= 0 && loopIndex < loops.get(classIndex).get(methIndex).size()))
			throw new ParseTreeIndexOutOfBoundsException("Illegal class, method or loop selected");
		com.sun.tools.javac.util.List<JmlTree.JmlStatementLoop> specs =
			getLoopSpecs(getLoop(classIndex,methIndex,loopIndex));
		return specs != null && !specs.isEmpty();
	}
	
	/** Replace the annotation of a method in the source code.
	 * 
	 * @param classIndex Index of the class to replace the annotation in
	 * @param methIndex Index of the method to replace the annotation in
	 * @param annotation The annotation to insert
	 * @return The original source code, with the selected annotation replaced
	 */
	public String replaceAnnotation(String code, int classIndex, int methIndex,
			com.sun.tools.javac.util.List<JmlSpecificationCase> annotation) {
		
		String adapted = code;
		int pos = getMethodPosition(classIndex,methIndex);
		
		//remove existing annotation
		if (isAnnotated(classIndex,methIndex)) {
			int annpos = methods.get(classIndex).get(methIndex).cases.cases.get(0).modifiers.pos;
			if (annpos < 0) //there are no modifiers
				annpos = methods.get(classIndex).get(methIndex).cases.pos;
			
			if (adapted.substring(0,annpos).trim().endsWith("//@"))
				annpos = adapted.substring(0,annpos).trim().lastIndexOf("//@");
			else if (adapted.substring(0,annpos).trim().endsWith("/*@"))
				annpos = adapted.substring(0,annpos).trim().lastIndexOf("/*@");
				
			adapted = adapted.substring(0,annpos) + adapted.substring(pos);
			pos = annpos;
		}
		
		//find the indentation
		String indent = adapted.substring(adapted.substring(0,pos).lastIndexOf('\n')+1, pos);
		
		//annotation to string
		String ann = "/*@\n" + indent 
			+ annotation.toString().replace("\n", "\n" + indent)
			+ "\n" + indent + " */\n" + indent;
		
		//add new annotation
		adapted = adapted.substring(0,pos) + ann + adapted.substring(pos);
		
		return adapted;
	}
	
	/** Replace the annotation of a loop in the source code.
	 * 
	 * @param classIndex Index of the class to replace the annotation in
	 * @param methIndex Index of the method to replace the annotation in
	 * @param loopIndex Index of the loop to replace the annotation in
	 * @param annotation The annotation to insert
	 * @return The original source code, with the selected annotation replaced
	 */
	public String replaceAnnotation(String code, int classIndex, int methIndex,
			int loopIndex, com.sun.tools.javac.util.List<JmlTree.JmlStatementLoop> annotation) {
		
		String adapted = code;
		int pos = getLoopPosition(classIndex,methIndex,loopIndex);
		
		//remove existing annotation
		if (isAnnotated(classIndex,methIndex,loopIndex)) {
			int annpos = getLoopSpecs(getLoop(classIndex,methIndex,loopIndex)).head.getStartPosition();
			
			if (adapted.substring(0,annpos).trim().endsWith("//@"))
				annpos = adapted.substring(0,annpos).trim().lastIndexOf("//@");
			else if (adapted.substring(0,annpos).trim().endsWith("/*@"))
				annpos = adapted.substring(0,annpos).trim().lastIndexOf("/*@");
			
			adapted = adapted.substring(0,annpos) + adapted.substring(pos);
			pos = annpos;
		}
		
		//find the indentation
		String indent = adapted.substring(adapted.substring(0,pos).lastIndexOf('\n')+1, pos);
		
		//annotation to string
		String ann = annotation.toString(",,,,,").replace(",,,,,", "\n"+indent) + "\n" + indent;
		
		//add new annotation
		adapted = adapted.substring(0,pos) + ann + adapted.substring(pos);
		
		return adapted;
	}
	
	/** Replace the annotation of loop and its parent method in the source code.
	 * 
	 * @param classIndex Index of the class to replace the annotation in
	 * @param methIndex Index of the method to replace the annotation in
	 * @param loopIndex
	 * @param methAnn The annotation to insert above the method
	 * @param loopAnn The annotation to insert above the loop
	 * @return The original source code, with the selected annotation replaced
	 */
	public String replaceAnnotations(String code, int classIndex, int methIndex, int loopIndex,
			com.sun.tools.javac.util.List<JmlSpecificationCase> methAnn,
			com.sun.tools.javac.util.List<JmlTree.JmlStatementLoop> loopAnn) {
		String newcode = replaceAnnotation(code, classIndex, methIndex, loopIndex, loopAnn);
		return replaceAnnotation(newcode, classIndex,methIndex,methAnn);
	}
	
	/** Get the OpenJML context
	 * 
	 * @return The OpenJML context
	 */
	public Context context() {
		return api.context();
	}
	
	/** Return the parse tree as a pretty printed string
	 * 
	 * @return A pretty-printed source representation of the parse tree
	 * @throws Exception
	 */
	public String prettyPrint() throws Exception {
		return api.prettyPrint(parsetree, true);
	}
	
	/** Get the exit condition from a loop.
	 * 
	 * @param loop The loop to get the exit-condition from
	 * @return The loop's exit condition (or null, if the supplied JCStatement was not a loop.
	 */
	public static JCExpression getLoopCondition(JCStatement loop) {
		if (loop instanceof JmlTree.JmlWhileLoop) {
			return ((JmlTree.JmlWhileLoop) loop).cond;
		} else if (loop instanceof JmlTree.JmlForLoop) {
			return ((JmlTree.JmlForLoop) loop).cond;
		} else if (loop instanceof JmlTree.JmlDoWhileLoop) {
			return ((JmlTree.JmlDoWhileLoop) loop).cond;
		} else {
				return null;
		}
	}
	
	/** Set the loop's exit-condition.
	 * 
	 * @param loop The loop to modify
	 * @param condition The new exit condition
	 * @return True on success, false if the supplied JCStatement was not a loop
	 */
	public static boolean setLoopCondition(JCStatement loop, JCExpression condition) {
		if (loop instanceof JmlTree.JmlWhileLoop) {
			((JmlTree.JmlWhileLoop) loop).cond = condition;
		} else if (loop instanceof JmlTree.JmlForLoop) {
			((JmlTree.JmlForLoop) loop).cond = condition;
		} else if (loop instanceof JmlTree.JmlDoWhileLoop) {
			((JmlTree.JmlDoWhileLoop) loop).cond = condition;
		} else {
				return false;
		}
		return true;
	}
	
	/** Get the body from a loop.
	 * 
	 * @param loop The loop to get the body from
	 * @return The loop's body (or null, if the supplied JCStatement was not a loop.
	 */
	public static JCStatement getLoopBody (JCStatement loop) {
		if (loop instanceof JCTree.JCWhileLoop)
			return ((JCTree.JCWhileLoop) loop).body;
		if (loop instanceof JCTree.JCForLoop)
			return ((JCTree.JCForLoop) loop).body;
		if (loop instanceof JCTree.JCDoWhileLoop)
			return ((JCTree.JCDoWhileLoop) loop).body;
		return null;
	}
	
	/** Set the loop's body
	 * 
	 * @param loop The loop to modify
	 * @param body The new body
	 * @return True on success, false if the supplied JCStatement was not a loop
	 */
	public static boolean setLoopBody(JCStatement loop, JCStatement body) {
		if (loop instanceof JmlTree.JmlWhileLoop) {
			((JmlTree.JmlWhileLoop) loop).body = body;
		} else if (loop instanceof JmlTree.JmlForLoop) {
			((JmlTree.JmlForLoop) loop).body = body;
		} else if (loop instanceof JmlTree.JmlDoWhileLoop) {
			((JmlTree.JmlDoWhileLoop) loop).body = body;
		} else {
				return false;
		}
		return true;
	}
		
	/** Get the JML specification from a loop.
	 * 
	 * @param loop The loop to get the JML specification from
	 * @return The loop's JML specification (or null, if the supplied JCStatement was not a loop.
	 */
	public static com.sun.tools.javac.util.List<JmlTree.JmlStatementLoop> getLoopSpecs (JCStatement loop) {
		if (loop instanceof JmlTree.JmlWhileLoop)
			return ((JmlTree.JmlWhileLoop) loop).loopSpecs;
		if (loop instanceof JCTree.JCForLoop)
			return ((JmlTree.JmlForLoop) loop).loopSpecs;
		if (loop instanceof JCTree.JCDoWhileLoop)
			return ((JmlTree.JmlDoWhileLoop) loop).loopSpecs;
		return null;
	}
	
	/** Set the loop's JML specification
	 * 
	 * @param loop The loop to modify
	 * @param specs The new JML specification
	 * @return True on success, false if the supplied JCStatement was not a loop
	 */
	public static boolean setLoopSpecs(JCStatement loop, com.sun.tools.javac.util.List<JmlTree.JmlStatementLoop> specs) {
		if (loop instanceof JmlTree.JmlWhileLoop) {
			((JmlTree.JmlWhileLoop) loop).loopSpecs = specs;
		} else if (loop instanceof JmlTree.JmlForLoop) {
			((JmlTree.JmlForLoop) loop).loopSpecs = specs;
		} else if (loop instanceof JmlTree.JmlDoWhileLoop) {
			((JmlTree.JmlDoWhileLoop) loop).loopSpecs = specs;
		} else {
				return false;
		}
		return true;
	}
	
	/** Removes the initialization statement (typically something like 'int i = 0')
	 * from a loop.
	 * 
	 * @param forloop The loop to modify
	 * @return True on success, false if the supplied JCStatement was not a loop
	 */
	public static boolean removeForLoopInit(JCStatement forloop) {
		if (forloop instanceof JmlTree.JmlForLoop) {
			((JmlTree.JmlForLoop) forloop).init = com.sun.tools.javac.util.List.<JCStatement> nil();
			return true;
		}
		return false;
	}
	
	/** Checks if a class of the given name exists.
	 * 
	 * @param className The classname to check for
	 * @return True, if a class of the given name exists
	 */
	public boolean classExists(String className) {
		for (JmlTree.JmlClassDecl c : classes) {
			if (c.name.toString().equals(className))
				return true;
		}
		return false;
	}
	
	/** Checks if a method of the given name exists in a class.
	 * 
	 * @param methodName The methodname to check for
	 * @param classIndex The class to look in
	 * @return True, if a method of the given name exists in the class
	 */
	public boolean methodExists(String methodName, int classIndex) {
		for (JmlTree.JmlMethodDecl m : methods.get(classIndex)) {
			if (m.name.toString().equals(methodName))
				return true;
		}
		return false;
	}
}
