package loopbounds.parsetree;

import loopbounds.LoopBoundsAnalyzer;
import loopbounds.VarInfo;

import java.util.List;
import java.util.ArrayList;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTreeScanner;
import org.jmlspecs.openjml.JmlTree.JmlVariableDecl;
import org.jmlspecs.openjml.JmlTree.JmlForLoop;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.*;
import com.sun.tools.javac.util.Name;

/**
 * Class for scanning the variables in an abstract syntax tree. The class is
 * intended to be used through the static functions 'containsVar' and
 * 'getUndeclaredVars'. The first checks whether a given variable name exists
 * in the given syntax tree. The second returns all variables in a given
 * syntax tree, which are not declared in that same tree.
 * <P>
 * The class extends JmlTreeScanner and uses the visitor pattern.
 * 
 * @author Martijn Moraal
 * @author Rody Kersten
 *
 */
public class VariableScanner extends JmlTreeScanner implements IJmlVisitor {
	
	private List<String> variables = new ArrayList<String>();
	private List<Name> varDecls = new ArrayList<Name>();
	
	/** Checks whether the given variables name exists in the
	 * given parse tree.
	 * 
	 * @param tree The parse tree to look in
	 * @param var The variable name to look for
	 * @return Whether or not the variable exists in the tree
	 */
	public static boolean containsVar (JCTree tree, Name var) {
		return containsVar(tree, var.toString());
	}
	
	/** Checks whether the given variables name exists in the
	 * given parse tree.
	 * 
	 * @param tree The parse tree to look in
	 * @param var The variable name to look for
	 * @return Whether or not the variable exists in the tree
	 */
	public static boolean containsVar (JCTree tree, String var) {	
		VariableScanner scanner = new VariableScanner();
		scanner.scan(tree);
		
		return containsName(var, scanner.varDecls)
			|| containsNameString(var, scanner.variables);
	}
	
	/** Gets a list of undeclared variables.
	 * 
	 * @param tree The parse tree to look in
	 * @return A list of the undeclared variables in 'tree'
	 */
	public static VarInfo getUndeclaredVars (JCTree tree) {
		VariableScanner scanner = new VariableScanner();
		scanner.scan(tree);
		
		List<String> undeclaredVars = new ArrayList<String>();
		for (int i = 0; i < scanner.variables.size(); i++) {
			if (!containsName(scanner.variables.get(i),scanner.varDecls)) {
				undeclaredVars.add(scanner.variables.get(i));
				if (LoopBoundsAnalyzer.DEBUG)
					System.out.println("Variable " + scanner.variables.get(i) +
							" is not declared in the given syntax tree.");
			}
		}
		
		return new VarInfo(undeclaredVars);
	}
	
	@Override
	public void visitIndexed(JCArrayAccess tree) {
		if (!variables.contains(tree.toString())) {
			variables.add(tree.toString());
		}
	}
	
	@Override
	public void visitIdent (JCIdent tree) {
 		if (!containsNameString(tree.name.toString(), variables) && !tree.name.toString().equalsIgnoreCase("this")) {
			variables.add(tree.toString());
		}
		super.visitIdent(tree);
	}
	
	@Override
	public void visitSelect(JCTree.JCFieldAccess field){
		if (!containsNameString(field.toString(), variables))
			variables.add(field.toString());
	}
	
	@Override
	public void visitJmlVariableDecl (JmlVariableDecl tree) {
		varDecls.add(tree.name);
		
		//don't scan the type, because if it's a class, that's an ident
        scan(tree.mods);
        scan(tree.init);
	}
	
	/* Ignore declaration within the for-loop definition */
	@Override
	public void visitJmlForLoop (JmlForLoop loop) {
		JmlForLoop copy = (JmlForLoop) loop.clone();
		copy.init = null;
		super.visitJmlForLoop(copy);
	}
		
	/* Ignore the invoked method's name (which is an 'ident') */
	@Override
    public void visitApply(JCMethodInvocation tree) {
        scan(tree.args);
    }
	
	/* Ignore the invoked method's name (which is an 'ident') */
	@Override
    public void visitNewClass(JCNewClass tree) {
		scan(tree.def);
    }
	
	/* Check wether a list contains a certain Name.
	 * @param name The name to look for
	 * @param lst The list to look in
	 * @return True if lst contains name, where character cases are ignored
	 */
	private static boolean containsName(String name, List<Name> lst) {
		for (Name n : lst) {
			if (n.toString().equalsIgnoreCase(name))
				return true;
		}
		return false;
	}
	
	/* Check wether a list contains a certain Name.
	 * @param name The name to look for
	 * @param lst The list to look in
	 * @return True if lst contains name, where character cases are ignored
	 */
	private static boolean containsNameString(String name, List<String> lst) {
		for (String n : lst) {
			if (n.equalsIgnoreCase(name))
				return true;
		}
		return false;
	}
}
