package resanaplugin;

import java.io.File;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaModelMarker;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.Signature;
import org.eclipse.jdt.internal.ui.actions.SelectionConverter;
import org.eclipse.jdt.internal.ui.javaeditor.JavaEditor;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.widgets.Shell;

@SuppressWarnings("restriction")
public class ProjectInformation {
	private JavaEditor activeEditor;
	private ISelection currentSelection;

	public ProjectInformation() {
		this.activeEditor = null;
		this.currentSelection = null;
	}

	public ProjectInformation(JavaEditor activeEditor,
			ISelection currentSelection) {
		this.activeEditor = activeEditor;
		this.currentSelection = currentSelection;
	}

	public boolean save(Shell shell) {
		if (activeEditor.isDirty()) {
			String name = activeEditor.getTitle();
			boolean save = MessageDialog.openQuestion(shell, "Save " + name,
					"Do you want to save " + name);
			if (save) {
				activeEditor.doSave(null);
			}
			return save;
		}
		return true;
	}

	public void changeSelection(ISelection newSelection) {
		currentSelection = newSelection;
	}

	public void setActiveEditor(JavaEditor newActiveEditor) {
		activeEditor = newActiveEditor;
	}

	public ICompilationUnit getCurrentFile() {
		return SelectionConverter.getInputAsCompilationUnit(activeEditor);
	}

	public IJavaProject getProject() {
		return getSelection().getJavaProject();
	}

	public IJavaElement getSelection() {
		if (currentSelection instanceof ITextSelection) {
			try {
				return SelectionConverter.resolveEnclosingElement(activeEditor,
						(ITextSelection) currentSelection);
			} catch (JavaModelException e) {
				e.printStackTrace();
			}
		}
		return null;
	}

	public boolean errors() {
		try {
			IMarker[] problems = getProject().getProject().findMarkers(
					IJavaModelMarker.JAVA_MODEL_PROBLEM_MARKER, true,
					IResource.DEPTH_INFINITE);
			for(IMarker problem : problems){
				if((Integer) problem.getAttribute("severity") == IMarker.SEVERITY_ERROR)
					return true;
			}
		} catch (CoreException e) {
			e.printStackTrace();
		}
		return false;
	}

	public static String getClassDeclaration(ICompilationUnit currentFile) {
		String classDeclaration = "";
		try {
			if (currentFile.getPackageDeclarations().length > 0) {
				classDeclaration = currentFile.getPackageDeclarations()[0]
						.getElementName() + "/";
				classDeclaration = classDeclaration.replace('.', '/');
			}
			classDeclaration += currentFile.getElementName().replace(".java",
					"");

		} catch (JavaModelException e1) {
			e1.printStackTrace();
		}
		return classDeclaration;
	}

	public static List<Method> getMethods(ICompilationUnit comp) {
		List<Method> methods = new LinkedList<Method>();
		try {
			for (IType type : comp.getAllTypes()) {
				for (IMethod method : type.getMethods()) {
					String returnTypeSignature = method.getReturnType();
					String returnType = Signature.toString(returnTypeSignature);
					String[][] t = type.resolveType(returnType);
					if (t != null) {
						returnTypeSignature = Signature.createTypeSignature(
								Signature.toQualifiedName(t[0]), true)
								.replaceAll("\\.", "/");
					}
					methods.add(new Method(method, method.getElementName()
							+ Signature.createMethodSignature(
									method.getParameterTypes(),
									returnTypeSignature)));
				}
			}
		} catch (JavaModelException e) {
			e.printStackTrace();
		}
		return methods;
	}

	public static String getClassPath(IJavaProject project) {
		try {
			IPath workingDirectory = ResourcesPlugin.getWorkspace().getRoot()
					.getProject(project.getElementName()).getLocation()
					.removeLastSegments(1);
			StringBuilder result = new StringBuilder();
			IClasspathEntry[] classpathEntries = project
					.getResolvedClasspath(false);
			for (IClasspathEntry classpathEntry : classpathEntries) {
				IPath path = null;
				switch (classpathEntry.getEntryKind()) {
				case IClasspathEntry.CPE_SOURCE:
					path = workingDirectory.append(classpathEntry.getPath());
					path = path.makeAbsolute();
					String binpath = path.toOSString().replace("src", "bin");
					result.append(binpath);
					result.append(File.pathSeparator);
					break;
				default:
					path = classpathEntry.getPath();
					path = path.makeAbsolute();
					break;
				}
				result.append(path.toOSString());
				result.append(File.pathSeparator);
			}
			return result.substring(0, result.length() - 1).toString();
		} catch (JavaModelException e) {
			e.printStackTrace();
		}
		return null;
	}

	public static String[] getClassPathArray(IJavaProject project) {
		try {
			IPath workingDirectory = ResourcesPlugin.getWorkspace().getRoot()
					.getProject(project.getElementName()).getLocation()
					.removeLastSegments(1);
			List<String> result = new ArrayList<String>();
			IClasspathEntry[] classpathEntries = project
					.getResolvedClasspath(false);
			for (IClasspathEntry classpathEntry : classpathEntries) {
				IPath path = null;
				switch (classpathEntry.getEntryKind()) {
				case IClasspathEntry.CPE_SOURCE:
					path = workingDirectory.append(classpathEntry.getPath());
					path = path.makeAbsolute();
					String binpath = path.toOSString().replace("src", "bin");
					result.add(binpath);
					break;
				default:
					path = classpathEntry.getPath();
					path = path.makeAbsolute();
					break;
				}
				result.add(path.toOSString());
			}
			return result.toArray(new String[result.size()]);
		} catch (JavaModelException e) {
			e.printStackTrace();
		}
		return null;
	}

	public ISourceViewer getViewer() {
		return activeEditor.getViewer();
	}
}
